#################################################################################
# Partitioned dynamic Bayesian networks
# 
# Author: Marcos L. de Paula Bueno
# marcoslbueno@gmail.com
# Last Update: May 14, 2016
#################################################################################

source("dbn-randbn.R")
source("dbn-randtrans.R")

#################################################################################
# Unrolling the structure of the PDBN
#################################################################################
structure.unrolled.tbn <- function (nodes, finals, dbn_init, dbns_paired) {
  allnames = c(sapply(0:(finals[length(finals)] - 1), function(x) { paste0(nodes,"_",x)} ))
  bn = empty.graph(allnames)
  
  #getting arcs
  arcs = dbn_init$arcs
  if (length(arcs) != 0) {
    for (i in 1:length(arcs[,1])) { #i=1
      arc = arcs[i,]
      bn = set.arc(bn, from = arc[1], to = arc[2])
    }
  }
  
  begin = 1
  end = finals[1]
  for (i in 1:length(dbns_paired)) { #i=1
    bn = arcs.dbn (bn, dbns_paired[[i]], begin, end)
    begin = end
    end = finals[i+1]
  }
  
  bn
}

arcs.dbn <- function (bn, dbn, begin, end) {
  arcs = dbn$arcs
  if (length(arcs) != 0) {
    for (i in 1:length(arcs[,1])) { #i=6
      for (j in begin:(end-1)) { #j=1
        arc = arcs[i, ]
        arc = gsub(x=arc, pattern = "_t2", replacement = paste0("_",j), fixed = T)
        arc = gsub(x=arc, pattern = "_t1", replacement = paste0("_",j-1), fixed = T)
        bn = set.arc(bn, from = arc[1], to = arc[2])
      }
    }
  }
  
  bn
}

#################################################################################
# Unrolling the parameters of the PDBN
#################################################################################
fit.unrolled.tbn <- function (bn, nodes, finals, fit_dbn_init, fits_paired) {
  numpoints = finals[length(finals)]
  fit = list()
  
  for (i in 1:length(fit_dbn_init)) { # i=3
    n = fit_dbn_init[[i]]
    
    if (length(n$parents) == 0) {
      data = n$prob
      m = matrix(data, ncol = length(data), dimnames = list(NULL, dimnames(data)[[1]]))
      dimnames(m)
    } else {
      data = c(n$prob)
      #Uniform replacement!
      data = as.numeric(gsub(pattern = NaN, replacement = 0.5, x = data))
      
      m = matrix(data, ncol = 1)
      dim(m) = dim(n$prob)
      dimnames(m) = dimnames(n$prob) 
    }
    
    fit[[i]] = m
  }
  
  begin = 1
  end = finals[1]
  for (i in 1:length(fits_paired)) { # i=2
    fit = cpt.tbn (fit, fits_paired[[i]], nodes, begin, end)
    begin = end
    end = finals[i+1]
  }
  
  allnames = paste0(nodes, "_0")
  
  begin = 1
  end = finals[1]
  aux = c()
  for (i in 1:length(fits_paired)) {    # i=2
    aux = c(sapply (nodes, function(x) { paste0(x, "_", begin:(end-1)) } ))
    allnames = c(allnames, aux)
    
    begin = end
    end = finals[i+1]
  }
  
  names(fit) = allnames
  
  f = custom.fit(bn, dist = fit)
  f
}

cpt.tbn <- function (fit, fit_dbn, nodes, begin, end) {
  
  base = length(nodes)
  
  for (i in 1:length(nodes)) { #i=1
    for (j in begin:(end-1)) { #j=1
      n = fit_dbn[[base + i]]
      
      if (length(n$parents) == 0) {
        data = n$prob
        m = matrix(data, ncol = length(data), dimnames = list(NULL, dimnames(data)[[1]]))
      } else {
        # renaming
        names(dimnames(n$prob)) = gsub (x=names(dimnames(n$prob)), pattern = "_t1", 
                                        replacement = paste0("_",j-1), fixed = T)
        names(dimnames(n$prob)) = gsub (x=names(dimnames(n$prob)), pattern = "_t2", 
                                        replacement = paste0("_",j), fixed = T)
        
        data = c(n$prob)
        #Uniform replacement for missing probabilities
        data = as.numeric(gsub(pattern = NaN, replacement = 0.5, x = data))
        
        m = matrix(data, ncol = 1)
        dim(m) = dim(n$prob)
        dimnames(m) = dimnames(n$prob) 
      }
      
      fit[[length(fit) + 1]] = m
    }
  }
  
  fit
}

#################################################################################
# Simulating from PDBNs
#################################################################################
dbn.simulate <- function (f, nodes, n) {
  train_simul = vector()
  shift = length(nodes) #sempre de 3 em 3
  
  for (k in 1:length(n)) { #k=1
    data = rbn(f, 1)
    
    seqsize = n[k]
    begin = 1
    for (i in 1:seqsize) { # i=2
      row = as.vector(t( data[begin:(begin + shift - 1)] ) )
      train_simul = rbind (train_simul, row)
      
      begin = begin+shift
    }
  }
  
  dimnames(train_simul) = list(NULL, nodes)
  
  train = list(x = train_simul, N = n)
  train
}

#################################################################################
# Random PDBNs
#################################################################################
buildrand.unrolled.dbn <- function (domsizes, finals, pr_node_init = NULL, pr_trans_same = NULL, pr_trans_other = NULL, pr_node_within = NULL) {
  #1 - random initial distribution
  if (is.null(pr_node_init)) { pr_node_init = 0.3 }
  
  dbni_structure = rand.structure.bn(numvar = length(domsizes), prob.node = pr_node_init)
  nodes = nodes(dbni_structure)
  nodes(dbni_structure) = paste0(nodes(dbni_structure),"_0")
  
  dbni = buildrand.bn(domsizes = domsizes, g = dbni_structure)
  # graphviz.plot(dbni$graph)
  
  #2 - random transition distributions
  dbns = vector(mode = "list", length = length(finals))
  fits = vector(mode = "list", length = length(finals))
  for (i in 1:length(finals)) {
    dbn = buildrand.dbn (domsizes = domsizes, pr_trans_same, pr_trans_other, pr_node_within)
    dbns[[i]] = dbn$graph
    fits[[i]] = dbn$fit
  }
  # graphviz.plot(dbns[[2]])
  
  #3 - unrolling structure
  bn = structure.unrolled.tbn(nodes, finals, dbni$graph, dbns)
  #graphviz.plot(bn)
  
  #4 - unrolling fit
  f = fit.unrolled.tbn(bn, nodes, finals, dbni$fit, fits)
  
  list(net = bn, fit = f, dbn_init = dbni, nets_paired = dbns, fits_paired = fits, train_paired = NULL)
}
